#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <wfdb/wfdb.h>
#include <wfdb/ecgmap.h>


#define MAX_LEADS 12
#define MAX_SIGS  5
#define MAX_BUF   100000
#define MAX_GAP   1000     

static WFDB_Siginfo s[MAX_LEADS];

unsigned stklen = 40000;
long nrSamps;
double *sig0;
double *sig1;
double *sig2;
double *sig3;
double *sig4;
double *sig5;
double *bloodPressure;

int m;
double dsum, dmin, dmax;
int openLeads;
long samples;
double normCnst = 32;

// dobim maximum na nekem podanem intervalu
// uporabljam za pridobitev prvega zacetnega thresholda
double getIntervalMax(start, stop) {
  double localMax = sig2[start];

  for(; start<=stop; start++) {
    if (sig2[start]>localMax) {
      localMax = sig2[start];
    }
  }

  return localMax;
}
// pridobitev indexa prvega lokalnega maximuma,
// uporabljam za pridobitev indexa maximuma potem v nadaljevanju
long getIntervalMaxIndex(start, stop) {
  double localMax = sig2[start];
  long maxIndex = start;

  for(; start<=stop; start++) {
    if (sig2[start]>localMax) {
      localMax = sig2[start];
      maxIndex = start;
    }
  }

  return maxIndex;
}

double calculateFirstThreshold() {
  // vzamem prvih 10 sekund in drugih 5 sekund - poiscem lokalni maksimum
  // obeh odsekov in ga povprecim ter izracunam po neki podani formuli za thresh
  // iz clanka
  long timeInterval = 5*250;
  double m1 = getIntervalMax(0, timeInterval); // max v prvih 5 sec
  double m2 = getIntervalMax(timeInterval, 2*timeInterval); // max v drugih 5 sec

  return (m1+m2)*0.5*0.5; // izracunam povprecje in vrnem polovico te vrednosti
}

double nonLinearLowPassFilter(long n) {
  long slidingWindow = 38; // sirina sliding windowa. 
  // taksno sirino predlaga avtor, ker je sampling rate 250Hz za EKG. 
  // in to korespondira z 150ms - tocna stevila je 37.5 ampak zaokrozimo navzgor
  double yn = 0;
  long i=0;
  // od trenutnega n-ja za velikost okna naprej, dokler ne dosezemo 
  // konca okna ali pa je prej konec signala
  for(i=n;i<(n+slidingWindow) && i<nrSamps; i++) {
    yn += sig1[i]*sig1[i]; // po formuli
  }

  return yn;
}

//prva stopnja po Chen-u. Izvedemo HighPass filter nad signalom
//M-point moving average filter. 
//formula:
//y1[n] = 1/M * SUM(x[n-m])
//
double highPassFilter(long n) {
  long m; // po njem sumiramo za usak signal posebej
  long M=7; //izbrano po predlogu clanka  M=7 ali M=5 to so priporocene vrednosti- Filter length
  double y1 = 0.0;
  double y2 = 0.0;

  if(n-M>=0) {  // pri prvih M vzorcih ne moremo izracunati y1 - zakasnitev
    for(m=0; m<M; m++) {
      y1 += sig0[n-m]; // po formuli izracunamo SUM-o
    }
  }
  y1 = y1 / M; // delimo rezultat z M - prvi del formule

  // izracunamo se drugi del HighPass Filtra y2 po formuli
  // y2[n] = x[n- (M+1)/2]
  if(n - ((M+1)/2) >= 0) { // pri 0-tem n-ju dobimo negativno stevilko...
    y2 = sig0[n-((M+1)/2)]; // po formuli
  } 

  // rezultat signala v tej tocki je enak po definiciji iz clanka
  // y[n] = y2[n] - y1[n];
  return y2-y1;
}

int findPeaks(double *source, double *writeTarget) {
  // poiscemo kandidate za QRS komplekse
  // iteriramo cez cel signal z drsecim oknom, najdemo maximum na na tem obmocju
  // in ga primerjamo z prejsnjim maximumom - na mesto maximuma shranimo 1, ce ga naslednji maximum 
  // v naslednjem oknu preseze ponastavimo prejsno vrednost na 0, ker je to nov maximum.
  // ponastavitev nazaj dela samo za en korak - je ok ker je okno 38samplov =150ms, dva qrs-a pa ne
  // moreta biti v tako ozkem itervalu
  long i;
  long indexOfLocalMax=0;
  long indexOfPreviousLocalMax=0;
  for(i=0; i<nrSamps; i+=38) {
    indexOfLocalMax = getIntervalMaxIndex(i, i+38); // poiscem index maximuma v tem intervalu
    if (source[indexOfLocalMax] > source[indexOfPreviousLocalMax]) { //primerjam z prejsnjim oknom
      writeTarget[indexOfLocalMax] = 1.0; // popravim vrednosti ce je novi max vecji
      writeTarget[indexOfPreviousLocalMax] = 0.0;
    }
    indexOfPreviousLocalMax = indexOfLocalMax; // zamenjam vrednost prejsnjega indexa
  }
  return 1;
}

double fixThreshold(current, peak) {
  // iz clanka pobrane vrednosti
  double alpha = 0.05;
  double gama = 0.15;
  return (alpha*gama*peak + (1-alpha)*current);
}

double detectQRS(int detCh){
  long i;
  // najprej izvedemo HighPass Filter
  fprintf(stderr,"Starting HighPass\n");
  for(i=0; i<nrSamps; i++) {
    sig1[i] = highPassFilter(i);
  }
  fprintf(stderr,"HighPass Completed\n");
  // izhod HighPass filtra nato po postopku spustimo v 
  // Nonlinear Low-Pass Filter
  // ki se izracuna - y[n] = SUM((y[n])^2) za neko doloceno okno
  fprintf(stderr,"Starting nonLinearLowPass\n");
  for(i=0; i<nrSamps; i++) {
    sig2[i] = nonLinearLowPassFilter(i);
  }
  fprintf(stderr,"nonLinearLowPass completed\n");

  // zdaj imamo sfiltriran signal - tako low-pass kot high-pass
  // zadnja stopnja algoritma je Decision-Making
  // dobimo prvi threshold na podlagi prvih 10 sekund signala - 2 maximuma povprecimo
  double threshold = calculateFirstThreshold();
  fprintf(stderr, "First threshold %f\n", threshold);

  double alpha = 0.05;
  double gama = 0.15;
  findPeaks(sig2, sig4);
  
  // najprej preveri ce je tocka kandidat - prejsnji korak findPeaks - ki nam je oznacil maximume
  // nato ce zadosti temu pogoju preveri tudi ce dosega treshold. Prejsnji korak nam dela torej
  // tudi preskoke naprej po oknu, kajti ce bi tukaj iskali ce presega tresh bi morali pogledati naprej
  // ce je v naslednjem oknu kaksen visji vrh. 
  // ce ne presega tresholda nasstavi na 0, kandidat ni QRS
  for(i=0; i<nrSamps;i++) {
    if (sig4[i] == 1.0 && sig2[i] > threshold) {
      threshold = alpha * gama * sig2[i] + (1 - alpha) * threshold;// popravi treshold po clanku
    }else {
      sig4[i] = 0.0;
    }
  }


  findPeaks(bloodPressure, sig5);

  long qrspeak = 0;
  long bppeak = 0;
  long qch = 0;
  long bch = 0;
  long firstqrs = 0;

  long averageBPDelay = 0;
  long matches = 0;

  for(i=0; i<nrSamps; i++) {
    if(sig4[i] == 1.0) {
      qrspeak = i;
      qch = 1;
      firstqrs = 1;
    }
    if(sig5[i] == 1.0 && firstqrs == 1) {
      bppeak = i;
      bch = 1;
    }

    if(qch == 1 && bch == 1) {
      fprintf(stderr, "Razdalja med BP peak in QRS peak %li \n", qrspeak - bppeak);
      qch=0;
      bch=0;
      averageBPDelay = averageBPDelay + abs(qrspeak-bppeak);
      matches = matches + 1;
    }

  }
  averageBPDelay = averageBPDelay/matches;
  fprintf(stderr, "Average BP Delay %li \n", averageBPDelay);

  long k = 0;
  long bpPeakFound=0;

  for(i=0; i<nrSamps; i++) {
    if(sig4[i] == 1.0) {
        for(k=i;k<i+averageBPDelay+38;k++){
          if(sig5[k]== 1.0) {
            bpPeakFound =1;
            break;
          }
        }
        if(bpPeakFound!=1){
          sig4[i] = 0;
        }else {
          i = i+averageBPDelay;
        }
    }
  }

  return threshold;
}

int
openRecord(char *record){
  return (isigopen(record, s, MAX_LEADS));
}

long ReadBuffer(long annot){
  long i;
  WFDB_Sample vec[MAX_LEADS];

  for (i=0;i<nrSamps;i++){
    if (getvec(vec)<0)
      break;
    sig0[i]=vec[0];
    bloodPressure[i]=vec[1];
  }
  return(nrSamps);
}

void 
writeQRS(char *record, int chan, char *ann){
  WFDB_Anninfo annIFO;
  WFDB_Annotation annot;
  int i;

  annIFO.name = ann; annIFO.stat = WFDB_WRITE;
  if (annopen(record, &annIFO, 1) < 0){
    fprintf(stderr,"Error opening QRS file\n");
    return;
  }
  annot.subtyp = annot.chan = annot.num = 0; annot.aux = NULL;
  for (i=0;i<samples;i++){
    if (sig4[i]!=0){
      annot.anntyp = sig4[i];
      annot.time=i;
      if (putann(0, &annot) < 0) break;
    }
  }
}

void 
readQRS(char *record, char *ant, int chan){
  WFDB_Anninfo annIFO;
  WFDB_Annotation annot;
  long i;

  annIFO.name = ant; annIFO.stat = WFDB_READ;
  if (annopen(record, &annIFO, 1) < 0){
    fprintf(stderr,"Error opening QRS file\n");
    return;
  }
  annot.subtyp = annot.chan = annot.num = 0; annot.aux = NULL;
  for (i=0;i<samples;i++) sig4[i]=0;
  i=0;
  while (getann(0, &annot)==0) {
    if (annot.time>samples) {
      fprintf(stderr,"Error reading annotation times\n");
      return;
    }
    sig4[annot.time]=annot.anntyp;
  }
}

int
main(int argc, char *argv[]){
  long i;
  char *record = NULL;
  char *annotator=NULL;
  double thresh;
  
  for (i=1;i<argc;i++){
    if (argv[i][0]!='-'){
      fprintf(stderr,"Error parsing command line\n");
      exit(1);
    }
    switch (argv[i][1]){
    case 'r':
      i++;
      record = (char *)malloc(sizeof(char)*(strlen(argv[i])+2));
      strcpy(record,argv[i]);
      break;
    case 'a': 
      i++;
      annotator = (char *) malloc(sizeof(char)*(strlen(argv[i])+2));
      strcpy(annotator, argv[i]);
      break;
    case 'n': 
      i++;
      normCnst = atof(argv[i]);
      break;
    default:
      fprintf(stderr,"Wrong switch\n");
      exit (2);
    }
  }
  if (record == NULL){
    fprintf(stderr,"No record was specified, exiting\n");
    exit (2);
  }
  if ((openLeads=openRecord(record))<0){
    fprintf(stderr,"Error opening record, exiting\n");
    exit (3);
  }

  fprintf(stderr,"Record opened\n");
  nrSamps = s->nsamp-1;
  sig0 = (double *)malloc(sizeof(double ) * nrSamps);
  sig1 = (double *)malloc(sizeof(double ) * nrSamps);
  sig2 = (double *)malloc(sizeof(double ) * nrSamps);
  sig3 = (double *)malloc(sizeof(double ) * nrSamps);
  sig4 = (double *)malloc(sizeof(double ) * nrSamps);
  sig5 = (double *)malloc(sizeof(double ) * nrSamps);
  bloodPressure = (double *)malloc(sizeof(double ) * nrSamps);
  if ((samples=ReadBuffer(0))<=0){
    fprintf(stderr,"Error opening record, exiting\n");
    exit (3);
  }
  fprintf(stderr,"Data read\n");
  if (samples!=nrSamps)
    fprintf(stderr,"Sample count differs\n");

  thresh=detectQRS(0);
  fprintf(stderr,"Data analyzed\n");
  if (annotator==NULL){
    annotator=(char *)malloc(sizeof(char)*5);
    sprintf(annotator,"qrs");
  }
  writeQRS(record, 4, annotator);
  fprintf(stderr,"Annotations written\n");
  wfdbquit();
  fprintf(stderr,"Record closed\n");
  free (sig0);
  free (sig1);
  free (sig2);
  free (sig3);
  free (sig4);
  free (sig5);
  free (bloodPressure);
  free (annotator);
  return 0;
}
